from collections import deque
from typing import Deque, Iterator, List, Dict
from enum import Enum

DEFAULT_WINDOW_SIZE = 40


class ChatMemoryEntry:
    """Represents a message in a conversation memory."""

    class Role(Enum):
        USER = "user"
        AGENT = "agent"

    def __init__(self, text: str, role: Role):
        self.text = text
        self.role = role


class ChatMemory:
    """Represents a sliding-window conversation memory between a user and an agent. `ChatMemory` is
    the most basic type of memory and is designed to be passed directly to chat-based models such
    as `llama3-70b-instruct`.
    """

    def __init__(self, window_size: int = DEFAULT_WINDOW_SIZE) -> None:
        """Create a new ChatMemory object.

        Args:
            window_size (int, optional): The maximum number of messages to store.
                Defaults to DEFAULT_WINDOW_SIZE.

        Raises:
            ValueError: If `window_size` is less than or equal to 0.
        """
        if not window_size > 0:
            raise ValueError("window_size must be a positive integer.")

        self.window_size: int = window_size
        self.mem_window: Deque[ChatMemoryEntry] = deque(maxlen=window_size)

    def add_user_message(self, text: str) -> None:
        """Adds a user message to the memory.

        Args:
            text (str): The user message to add.
        """
        self.mem_window.append(ChatMemoryEntry(text, ChatMemoryEntry.Role.USER))

    def add_agent_message(self, text: str) -> None:
        """Adds an agent message to the memory.

        Args:
            text (str): The agent message to add.
        """
        self.mem_window.append(ChatMemoryEntry(text, ChatMemoryEntry.Role.AGENT))

    def unpack(
        self, role_key: str, message_key: str, user_key: str, agent_key: str
    ) -> List[Dict[str, str]]:
        """Gets a representation of the memory as a list of objects designed to
        be passed directly into LLM provider APIs as JSON.

        For example, with the following memory:
        ```
        memory = ChatMemory()
        memory.add_user_message("Hello")
        memory.add_agent_message("Hi!")
        memory.add_user_message("How are you?")
        memory.add_agent_message("I'm good, how are you?")
        ```
        `memory.unpack("role", "content", "user", "assistant")` would return:
        ```
        [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi!"},
            {"role": "user", "content": "How are you?"},
            {"role": "assistant", "content": "I'm good, how are you?"}
        ]
        ```

        Args:
            role_key (str): The key to use to denote role. For example, "role".
            message_key (str): The key to use to denote the message. For example, "content".
            user_key (str): The key to denote the user's message. For example, "user".
            agent_key (str): The key to denote the agent or model's message. For example, "assistant".

        Returns:
            List[Dict[str, str]]: The representation of the memory as a list of objects.
        """
        res = []
        for message in self.mem_window:
            role_value = (
                user_key if message.role == ChatMemoryEntry.Role.USER else agent_key
            )
            res.append({role_key: role_value, message_key: message.text})
        return res

    def clear(self) -> None:
        """Clears the memory."""
        self.mem_window.clear()

    def __len__(self) -> int:
        return len(self.mem_window)

    def __iter__(self) -> Iterator[ChatMemoryEntry]:
        return iter(self.mem_window)



from typing import Any, Set, Dict, Optional

import google.generativeai as google
from cohere import Client as CohereClient
from openai import OpenAI
from anthropic import Anthropic
from groq import Groq
import replicate

from l2m2.model_info import MODEL_INFO, PROVIDER_INFO, PROVIDER_DEFAULT
from l2m2.memory import ChatMemory, DEFAULT_WINDOW_SIZE


class LLMClient:
    """A high-level interface for interacting with L2M2's supported language models."""

    def __init__(
        self,
        providers: Optional[Dict[str, str]] = None,
        enable_memory: bool = False,
        memory_window_size: int = DEFAULT_WINDOW_SIZE,
    ) -> None:
        """Initialize the LLMClient, optionally with active providers.

        Args:
            providers ([Dict[str, str]], optional): Mapping from provider name to API key.
                For example::

                    {
                        "openai": "openai-api
                        "anthropic": "anthropic-api-key",
                        "google": "google-api-key",
                    }

                Defaults to `None`.
            enable_memory (bool, optional): Whether to enable memory. Defaults to `False`.
            window_size (int, optional): The size of the memory window. Defaults to
                `l2m2.memory.DEFAULT_WINDOW_SIZE`.

        Raises:
            ValueError: If an invalid provider is specified in `providers`.
            ValueError: If `memory_window_size` is not a positive integer.
        """
        self.api_keys: Dict[str, str] = {}
        self.active_providers: Set[str] = set()
        self.active_models: Set[str] = set()
        self.preferred_providers: Dict[str, str] = {}

        self.memory: Optional[ChatMemory] = None

        if providers is not None:
            for provider, api_key in providers.items():
                self.add_provider(provider, api_key)

        if enable_memory:
            self.memory = ChatMemory(memory_window_size)

    @staticmethod
    def get_available_providers() -> Set[str]:
        """Get the exhaustive set of L2M2's available model providers. This set includes
        all providers, regardless of whether they are currently active. L2M2 does not currently
        support adding custom providers.

        Returns:
            Set[str]: A set of available providers.
        """
        return set(PROVIDER_INFO.keys())

    @staticmethod
    def get_available_models() -> Set[str]:
        """The set of L2M2's supported models. This set includes all models, regardless of
        whether they are currently active. L2M2 allows users to call non-available models
        from available and active providers, but does not guarantee correctness for such calls.

        Returns:
            Set[str]: A set of available models.
        """
        return set(MODEL_INFO.keys())

    def get_active_providers(self) -> Set[str]:
        """Get the set of currently active providers. Active providers are those for which an API
        key has been set.

        Returns:
            Set[str]: A set of active providers.
        """
        return set(self.active_providers)

    def get_active_models(self) -> Set[str]:
        """Get the set of currently active models. Active models are those that are available and
        have a provider that is active.

        Returns:
            Set[str]: A set of active models.
        """
        return set(self.active_models)

    def add_provider(self, provider: str, api_key: str) -> None:
        """Add a provider to the LLMClient, making its models available for use.

        Args:
            provider (str): The provider name. Must be one of the available providers.
            api_key (str): The API key for the provider.

        Raises:
            ValueError: If the provider is not one of the available providers.
        """
        if provider not in (providers := self.get_available_providers()):
            raise ValueError(
                f"Invalid provider: {provider}. Available providers: {providers}"
            )

        self.api_keys[provider] = api_key
        self.active_providers.add(provider)
        self.active_models.update(
            model for model in MODEL_INFO.keys() if provider in MODEL_INFO[model].keys()
        )

    def remove_provider(self, provider_to_remove: str) -> None:
        """Remove a provider from the LLMClient, making its models unavailable for use.

        Args:
            provider (str): The active provider to remove.

        Raises:
            ValueError: If the given provider is not active.
        """
        if provider_to_remove not in self.active_providers:
            raise ValueError(f"Provider not active: {provider_to_remove}")

        del self.api_keys[provider_to_remove]
        self.active_providers.remove(provider_to_remove)

        self.active_models = {
            model
            for model in self.active_models
            if not MODEL_INFO[model].keys().isdisjoint(self.active_providers)
        }

    def set_preferred_providers(self, preferred_providers: Dict[str, str]) -> None:
        """Set the preferred provider for each model. If a model is available from multiple active
        providers, the preferred provider will be used.

        Args:
            preferred_providers (Dict[str, str]): A mapping from model name to preferred provider.
                For example::

                    {
                        "llama3-8b": "groq",
                        "llama3-70b": "replicate",
                    }

                If you'd like to remove a preferred provider, set it to `None`, e.g.::

                    {
                        "llama3-8b": None,
                    }

        Raises:
            ValueError: If an invalid model or provider is specified in `preferred_providers`.
            ValueError: If the given model does not correlate to the given provider.
        """
        for model, provider in preferred_providers.items():
            if model not in self.get_available_models():
                raise ValueError(f"Invalid model: {model}")
            if provider is not None:
                if provider not in self.get_available_providers():
                    raise ValueError(f"Invalid provider: {provider}")
                if provider not in MODEL_INFO[model].keys():
                    raise ValueError(
                        f"Model {model} is not available from provider {provider}."
                    )

        self.preferred_providers.update(preferred_providers)

    def get_memory(self) -> ChatMemory:
        """Get the memory object, if memory is enabled.

        Returns:
            ChatMemory: The memory object.

        Raises:
            ValueError: If memory is not enabled.
        """
        if self.memory is None:
            raise ValueError(
                "Client memory is not enabled. Instantiate the LLM client with enable_memory=True"
                + " to enable memory."
            )
        return self.memory

    def clear_memory(self) -> None:
        """Clear the memory, if memory is enabled.

        Raises:
            ValueError: If memory is not enabled.
        """
        if self.memory is None:
            raise ValueError(
                "Client memory is not enabled. Instantiate the LLM client with enable_memory=True"
                + " to enable memory."
            )
        self.memory.clear()

    def enable_memory(self, window_size: int = DEFAULT_WINDOW_SIZE) -> None:
        """Enable memory, with a specified window size.

        Args:
            window_size (int, optional): The size of the memory window. Defaults to
                `l2m2.memory.DEFAULT_WINDOW_SIZE`.

        Raises:
            ValueError: If memory is already enabled.
        """
        if self.memory is not None:
            raise ValueError("Memory is already enabled.")
        self.memory = ChatMemory(window_size)

    def call(
        self,
        *,
        model: str,
        prompt: str,
        system_prompt: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        prefer_provider: Optional[str] = None,
    ) -> str:
        """Performs inference on any active model.

        Args:
            model (str): The active model to call.
            prompt (str): The user prompt for which to generate a completion.
            system_prompt (str, optional): The system prompt to send to the model. If the specified
                model does not support system prompts, it is prepended to the user prompt. Defaults
                to None.
            temperature (float, optional): The sampling temperature for the model. If not specified,
                the provider's default value for the model is used. Defaults to None.
            max_tokens (int, optional): The maximum number of tokens to generate. If not specified,
                the provider's default value for the model is used. Defaults to None.
            prefer_provider (str, optional): The preferred provider to use for the model, if the
                model is available from multiple active providers. Defaults to None.

        Raises:
            ValueError: If the provided model is not active and/or not available.
            ValueError: If the model is available from multiple active providers neither `prefer_provider`
                nor a default provider is specified.
            ValueError: If `prefer_provider` is specified but not active.

        Returns:
            str: The model's completion for the prompt, or an error message if the model is
                unable to generate a completion.
        """
        if model not in self.active_models:
            if model in self.get_available_models():
                available_providers = ", ".join(MODEL_INFO[model].keys())
                msg = (
                    f"Model {model} is available, but not active."
                    + f" Please add any of ({available_providers}) to activate it."
                )
                raise ValueError(msg)
            else:
                raise ValueError(f"Invalid model: {model}")

        if prefer_provider is not None and prefer_provider not in self.active_providers:
            raise ValueError(
                "Argument prefer_provider must either be None or an active provider."
                + f" Active providers are {', '.join(self.active_providers)}"
            )

        providers = set(MODEL_INFO[model].keys()) & self.active_providers
        if len(providers) == 1:
            provider = next(iter(providers))

        elif prefer_provider is not None:
            provider = prefer_provider

        elif self.preferred_providers.get(model) is not None:
            provider = self.preferred_providers[model]

        else:
            raise ValueError(
                f"Model {model} is available from multiple active providers: {', '.join(providers)}."
                + " Please specify a preferred provider with the argument prefer_provider, or set a"
                + " default provider for the model with set_preferred_providers."
            )

        return self._call_impl(
            MODEL_INFO[model][provider],
            provider,
            prompt,
            system_prompt,
            temperature,
            max_tokens,
        )

    def call_custom(
        self,
        *,
        provider: str,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
    ) -> str:
        """Performs inference on any model from an active provider that is not officially supported
        by L2M2. This method does not guarantee correctness.

        Args:
            provider (str): The provider to use. Must be one of the active providers.
            model_id (str): The ID of model to call. Must be the exact match to how you would call
                it with the provider's API. For example, `gpt-3.5-turbo-0125` can be used to call
                a legacy model from OpenAI as per the OpenAI API docs.
                (https://platform.openai.com/docs/api-reference/chat)
            prompt (str): The user prompt for which to generate a completion.
            system_prompt (str, optional): The system prompt to send to the model. Defaults to None.
            temperature (float, optional): The sampling temperature for the model. If not specified,
                the provider's default value for the model is used. Defaults to None.
            max_tokens (int, optional): The maximum number of tokens to generate. If not specified,
                the provider's default value for the model is used. Defaults to None.

        Raises:
            ValueError: If the provided model is not active and/or not available.

        Returns:
            str: The model's completion for the prompt (correctness not guaranteed).
        """
        if provider not in self.get_available_providers():
            raise ValueError(f"Invalid provider: {provider}")
        if provider not in self.active_providers:
            raise ValueError(f"Provider not active: {provider}")

        # Get the param info from the first model where the provider matches.
        # Not ideal, but the best we can do for user-provided models.
        model_info = {
            "model_id": model_id,
            "params": MODEL_INFO[
                next(
                    model
                    for model in self.get_available_models()
                    if provider in MODEL_INFO[model].keys()
                )
            ][provider]["params"],
        }

        return self._call_impl(
            model_info,
            provider,
            prompt,
            system_prompt,
            temperature,
            max_tokens,
        )

    def _call_impl(
        self,
        model_info: Dict[str, Any],
        provider: str,
        prompt: str,
        system_prompt: Optional[str],
        temperature: Optional[float],
        max_tokens: Optional[int],
    ) -> str:
        param_info = model_info["params"]
        params = {}

        def add_param(name: str, value: Any) -> None:
            if value is not None and value > (max_val := param_info[name]["max"]):
                msg = f"Parameter {name} exceeds max value {max_val}"
                raise ValueError(msg)

            key = name
            if "custom_key" in param_info[name]:
                key = param_info[name]["custom_key"]

            if value is not None:
                params[key] = value
            elif (default := param_info[name]["default"]) != PROVIDER_DEFAULT:
                params[key] = default

        add_param("temperature", temperature)
        add_param("max_tokens", max_tokens)

        result = getattr(self, f"_call_{provider}")(
            model_info["model_id"], prompt, system_prompt, params
        )
        assert isinstance(result, str), "This should never happen."
        if self.memory is not None:
            self.memory.add_user_message(prompt)
            self.memory.add_agent_message(result)
        return result

    def _call_openai(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        oai = OpenAI(api_key=self.api_keys["openai"])
        messages = []
        if system_prompt is not None:
            messages.append({"role": "system", "content": system_prompt})
        if self.memory is not None:
            messages.extend(self.memory.unpack("role", "content", "user", "assistant"))
        messages.append({"role": "user", "content": prompt})
        result = oai.chat.completions.create(
            model=model_id,
            messages=messages,  # type: ignore
            **params,
        )
        return str(result.choices[0].message.content)

    def _call_anthropic(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        anthr = Anthropic(api_key=self.api_keys["anthropic"])
        if system_prompt is not None:
            params["system"] = system_prompt
        messages = []
        if self.memory is not None:
            messages.extend(self.memory.unpack("role", "content", "user", "assistant"))
        messages.append({"role": "user", "content": prompt})
        result = anthr.messages.create(
            model=model_id,
            messages=messages,  # type: ignore
            **params,
        )
        return str(result.content[0].text)

    def _call_cohere(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        cohere = CohereClient(api_key=self.api_keys["cohere"])
        if system_prompt is not None:
            params["preamble"] = system_prompt
        if self.memory is not None:
            params["chat_history"] = self.memory.unpack(
                "role", "message", "USER", "CHATBOT"
            )
        result = cohere.chat(
            model=model_id,
            message=prompt,
            **params,
        )
        return str(result.text)

    def _call_groq(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        groq = Groq(api_key=self.api_keys["groq"])
        messages = []
        if system_prompt is not None:
            messages.append({"role": "system", "content": system_prompt})
        if self.memory is not None:
            messages.extend(self.memory.unpack("role", "content", "user", "assistant"))
        messages.append({"role": "user", "content": prompt})
        result = groq.chat.completions.create(
            model=model_id,
            messages=messages,  # type: ignore
            **params,
        )
        return str(result.choices[0].message.content)

    def _call_google(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        google.configure(api_key=self.api_keys["google"])

        model_params = {"model_name": model_id}
        if system_prompt is not None:
            # Earlier versions don't support system prompts, so prepend it to the prompt
            if model_id not in ["gemini-1.5-pro-latest"]:
                prompt = f"{system_prompt}\n{prompt}"
            else:
                model_params["system_instruction"] = system_prompt
        model = google.GenerativeModel(**model_params)

        messages = []
        if self.memory is not None:
            messages.extend(self.memory.unpack("role", "parts", "user", "model"))
        messages.append({"role": "user", "parts": prompt})

        result = model.generate_content(messages, generation_config=params)
        result = result.candidates[0]

        # Will sometimes fail due to safety filters
        if result.content:
            return str(result.content.parts[0].text)
        else:
            return str(result)

    def _call_replicate(
        self,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str],
        params: Dict[str, Any],
    ) -> str:
        if self.memory is not None:
            raise ValueError("Memory is not supported with Replicate models.")
        client = replicate.Client(api_token=self.api_keys["replicate"])
        if system_prompt is not None:
            params["system_prompt"] = system_prompt
        result = client.run(
            model_id,
            input={
                "prompt": prompt,
                **params,
            },
        )
        return "".join(result)



"""Information about models and providers supported by L2M2."""

from typing import Any, Dict

PROVIDER_DEFAULT = "<<PROVIDER_DEFAULT>>"


PROVIDER_INFO = {
    "openai": {
        "name": "OpenAI",
        "homepage": "https://openai.com/product",
    },
    "google": {
        "name": "Google",
        "homepage": "https://ai.google.dev/",
    },
    "anthropic": {
        "name": "Anthropic",
        "homepage": "https://www.anthropic.com/api",
    },
    "cohere": {
        "name": "Cohere",
        "homepage": "https://docs.cohere.com/",
    },
    "groq": {
        "name": "Groq",
        "homepage": "https://wow.groq.com/",
    },
    "replicate": {
        "name": "Replicate",
        "homepage": "https://replicate.com/",
    },
}

MODEL_INFO: Dict[str, Any] = {
    "gpt-4-turbo": {
        "openai": {
            "model_id": "gpt-4-turbo-2024-04-09",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 4096,
                },
            },
        },
    },
    "gpt-3.5-turbo": {
        "openai": {
            "model_id": "gpt-3.5-turbo-0125",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 4096,
                },
            },
        },
    },
    "gemini-1.5-pro": {
        "google": {
            "model_id": "gemini-1.5-pro-latest",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "custom_key": "max_output_tokens",
                    "default": PROVIDER_DEFAULT,
                    # https://cloud.google.com/vertex-ai/generative-ai/docs/learn/models#gemini-models
                    "max": 8192,
                },
            },
        },
    },
    "gemini-1.0-pro": {
        "google": {
            "model_id": "gemini-1.0-pro-latest",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "custom_key": "max_output_tokens",
                    "default": PROVIDER_DEFAULT,
                    # https://cloud.google.com/vertex-ai/generative-ai/docs/learn/models#gemini-models
                    "max": 8192,
                },
            },
        },
    },
    "claude-3-opus": {
        "anthropic": {
            "model_id": "claude-3-opus-20240229",
            "params": {
                "temperature": {
                    "default": 0.0,
                    "max": 1.0,
                },
                "max_tokens": {
                    "default": 1000,  # L2M2 default, field is required
                    "max": 4096,
                },
            },
        },
    },
    "claude-3-sonnet": {
        "anthropic": {
            "model_id": "claude-3-sonnet-20240229",
            "params": {
                "temperature": {
                    "default": 0.0,
                    "max": 1.0,
                },
                "max_tokens": {
                    "default": 1000,  # L2M2 default, field is required
                    "max": 4096,
                },
            },
        },
    },
    "claude-3-haiku": {
        "anthropic": {
            "model_id": "claude-3-haiku-20240307",
            "params": {
                "temperature": {
                    "default": 0.0,
                    "max": 1.0,
                },
                "max_tokens": {
                    "default": 1000,  # L2M2 default, field is required
                    "max": 4096,
                },
            },
        },
    },
    "command-r": {
        "cohere": {
            "model_id": "command-r",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 1.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 4000,
                },
            },
        },
    },
    "command-r-plus": {
        "cohere": {
            "model_id": "command-r-plus",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 1.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 4000,
                },
            },
        },
    },
    "mixtral-8x7b": {
        "groq": {
            "model_id": "mixtral-8x7b-32768",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2**16 - 1,
                },
            },
        },
    },
    "gemma-7b": {
        "groq": {
            "model_id": "gemma-7b-it",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2.0,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2**16 - 1,
                },
            },
        },
    },
    "llama3-8b": {
        "groq": {
            "model_id": "llama3-8b-8192",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2**16 - 1,
                },
            },
        },
        "replicate": {
            "model_id": "meta/meta-llama-3-8b-instruct",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 5.0,
                },
                "max_tokens": {
                    "custom_key": "max_new_tokens",
                    "default": PROVIDER_DEFAULT,
                    "max": float("inf"),
                },
            },
        },
    },
    "llama3-70b": {
        "groq": {
            "model_id": "llama3-70b-8192",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2,
                },
                "max_tokens": {
                    "default": PROVIDER_DEFAULT,
                    "max": 2**16 - 1,
                },
            },
        },
        "replicate": {
            "model_id": "meta/meta-llama-3-8b-instruct",
            "params": {
                "temperature": {
                    "default": PROVIDER_DEFAULT,
                    "max": 5.0,
                },
                "max_tokens": {
                    "custom_key": "max_new_tokens",
                    "default": PROVIDER_DEFAULT,
                    "max": float("inf"),
                },
            },
        },
    },
}



import asyncio
import inspect
from typing import Optional, List, Any

from l2m2.client import LLMClient


class AsyncLLMClient(LLMClient):
    """A high-level interface for asynchronously interacting with L2M2's language models."""

    def from_client(client: LLMClient) -> "AsyncLLMClient":
        """Create an AsyncLLMClient from an LLMClient.

        Args:
            client (LLMClient): The base client to wrap.

        Returns:
            AsyncLLMClient: The new asynchronous client with the same models and providers active.
        """
        aclient = AsyncLLMClient(client.api_keys)
        aclient.set_preferred_providers(client.preferred_providers)
        return aclient

    async def call_async(
        self,
        *,
        model: str,
        prompt: str,
        system_prompt: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        prefer_provider: Optional[str] = None,
    ) -> str:
        """Asynchronously performs inference on any active model.

        Args:
            model (str): The active model to call.
            prompt (str): The user prompt for which to generate a completion.
            system_prompt (str, optional): The system prompt to send to the model. If the specified
                model does not support system prompts, it is prepended to the user prompt. Defaults
                to None.
            temperature (float, optional): The sampling temperature for the model. If not specified,
                the provider's default value for the model is used. Defaults to None.
            max_tokens (int, optional): The maximum number of tokens to generate. If not specified,
                the provider's default value for the model is used. Defaults to None.
            prefer_provider (str, optional): The preferred provider to use for the model, if the
                model is available from multiple active providers. Defaults to None.

        Raises:
            ValueError: If the provided model is not active and/or not available.
            ValueError: If the model is available from multiple active providers neither `prefer_provider`
                nor a default provider is specified.
            ValueError: If `prefer_provider` is specified but not active.

        Returns:
            str: The model's completion for the prompt, or an error message if the model is
                unable to generate a completion.
        """
        return await self._call_wrapper(
            model=model,
            prompt=prompt,
            system_prompt=system_prompt,
            temperature=temperature,
            max_tokens=max_tokens,
            prefer_provider=prefer_provider,
        )

    async def call_custom_async(
        self,
        *,
        provider: str,
        model_id: str,
        prompt: str,
        system_prompt: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
    ) -> str:
        """Asynchronously Performs inference on any model from an active provider that is not
        officially supported by L2M2. This method does not guarantee correctness.

        Args:
            provider (str): The provider to use. Must be one of the active providers.
            model_id (str): The ID of model to call. Must be the exact match to how you would call
                it with the provider's API. For example, `gpt-3.5-turbo-0125` can be used to call
                a legacy model from OpenAI as per the OpenAI API docs.
                (https://platform.openai.com/docs/api-reference/chat)
            prompt (str): The user prompt for which to generate a completion.
            system_prompt (str, optional): The system prompt to send to the model. Defaults to None.
            temperature (float, optional): The sampling temperature for the model. If not specified,
                the provider's default value for the model is used. Defaults to None.
            max_tokens (int, optional): The maximum number of tokens to generate. If not specified,
                the provider's default value for the model is used. Defaults to None.

        Raises:
            ValueError: If the provided model is not active and/or not available.

        Returns:
            str: The model's completion for the prompt (correctness not guaranteed).
        """
        return await self._call_custom_wrapper(
            provider=provider,
            model_id=model_id,
            prompt=prompt,
            system_prompt=system_prompt,
            temperature=temperature,
            max_tokens=max_tokens,
        )

    async def call_concurrent(
        self,
        *,
        n: int,
        models: List[str],
        prompts: List[str],
        system_prompts: Optional[List[str]] = None,
        temperatures: Optional[List[float]] = None,
        max_tokens: Optional[List[int]] = None,
        prefer_providers: Optional[List[str]] = None,
    ) -> List[str]:
        """Makes multiple concurrent calls to a given set of models, with a given set oof
        parameters (e.g. model, prompt, temperature, etc). Each parameter is passed in as a list
        of length 1 or n, where n is the number of calls to make. If a parameter is a list of
        length 1, the value is shared across all calls.

        For example, ::

            await client.call_concurrent(
                n=2,
                models=["gpt-4-turbo", "claude-3-opus"],
                prompts=["foo", "bar"],
            )

        will do send `foo` to `gpt-4-turbo` and `bar` to `claude-3-opus`, while ::

            await client.call_concurrent(
                n=2,
                models=["gpt-4-turbo", "claude-3-opus"],
                prompts=["foo"],
            )

        will send `foo` to both models. Similarly, ::

            await client.call_concurrent(
                n=2,
                models=["gpt-4-turbo"],
                prompts=["foo", "bar"],
            )

        will prompt `gpt-4-turbo` with both `foo` and `bar`.

        Args:
            n (int): Number of concurrent calls to make.
            models (List[str]): List of models to call, or a single model to use for all calls.
            prompts (List[str]): List of prompts to use, or a single prompt to use for all calls.
            system_prompts ([List[str]], optional): List of system prompts to use, or a single
                system prompt to use for all calls. Defaults to None.
            temperatures ([List[float]], optional): List of temperatures to use, or a single
                temperature to use for all calls. Defaults to None.
            max_tokens ([List[int]], optional): List of max_tokens to use, or a single max_tokens
                to use for all calls. Defaults to None.
            prefer_providers ([List[str]], optional): List of preferred providers to use for each
                model, if the model is available from multiple active providers. Defaults to None.

        Raises:
            ValueError: If `n < 1`, or if any of the parameters are not of length `1` or `n`.
            ValueError: If the provided model is not active and/or not available.
            ValueError: If the model is available from multiple active providers neither `prefer_provider`
                nor a default provider is specified.
            ValueError: If `prefer_provider` is specified but not active.

        Returns:
            List[str]: The responses from each call, in the same order as the input parameters.
        """
        _check_concurrent_params(
            n,
            [models, prompts, system_prompts, temperatures, max_tokens],
            inspect.getfullargspec(self.call_concurrent).kwonlyargs,
        )

        calls = [
            self._call_wrapper(
                model=_get_helper(models, i),
                prompt=_get_helper(prompts, i),
                system_prompt=_get_helper(system_prompts, i),
                temperature=_get_helper(temperatures, i),
                max_tokens=_get_helper(max_tokens, i),
                prefer_provider=_get_helper(prefer_providers, i),
            )
            for i in range(n)
        ]
        return await asyncio.gather(*calls)

    async def call_custom_concurrent(
        self,
        *,
        n: int,
        providers: List[str],
        model_ids: List[str],
        prompts: List[str],
        system_prompts: Optional[List[str]] = None,
        temperatures: Optional[List[float]] = None,
        max_tokens: Optional[List[int]] = None,
    ) -> List[str]:
        """Makes multiple concurrent calls to a given set of user-given models, with a given set oof
        parameters (e.g. model_d, prompt, temperature, etc). Each parameter is passed in as a list
        of length 1 or n, where n is the number of calls to make. If a parameter is a list of
        length 1, the value is shared across all calls.

        For example, ::

            await client.call_custom_concurrent(
                n=2,
                providers=["openai", "anthropic"],
                model_ids=["gpt-4-turbo", "claude-3-opus"],
                prompts=["foo", "bar"],
            )

        will do send `foo` to `gpt-4-turbo` and `bar` to `claude-3-opus`, while ::

            await client.call_custom_concurrent(
                n=2,
                providers=["openai", "anthropic"],
                model_ids=["gpt-4-turbo", "claude-3-opus"],
                prompts=["foo"],
            )

        will send `foo` to both models. Similarly, ::

            await client.call_custom_concurrent(
                n=2,
                providers=["openai"],
                model_ids=["gpt-4-turbo"],
                prompts=["foo", "bar"],
            )

        will prompt `gpt-4-turbo` with both `foo` and `bar`.

        Simmiarly to `call_custom_async`, `call_custom_concurrent` allows you to call models
        from active providers that are not officially supported by L2M2.
        `call_custom_concurrent` does not guarantee correctness.

        Args:
            n (int): Number of concurrent calls to make.
            providers (List[str]): List of providers to use, or a single provider to use for all calls.
            model_ids (List[str]): List of models to call, or a single model to use for all calls.
            prompts (List[str]): List of prompts to use, or a single prompt to use for all calls.
            system_prompts ([List[str]], optional): List of system prompts to use, or a single
                system prompt to use for all calls. Defaults to None.
            temperatures ([List[float]], optional): List of temperatures to use, or a single
                temperature to use for all calls. Defaults to None.
            max_tokens ([List[int]], optional): List of max_tokens to use, or a single max_tokens
                to use for all calls. Defaults to None.

        Raises:
            ValueError: If `n < 1`, or if any of the parameters are not of length `1` or `n`.

        Raises:
            ValueError: If the provided model is not active and/or not available.

        Returns:
            List[str]: The responses from each call, in the same order as the input parameters.
        """
        _check_concurrent_params(
            n,
            [providers, model_ids, prompts, system_prompts, temperatures, max_tokens],
            inspect.getfullargspec(self.call_custom_concurrent).kwonlyargs,
        )

        calls = [
            self._call_custom_wrapper(
                provider=_get_helper(providers, i),
                model_id=_get_helper(model_ids, i),
                prompt=_get_helper(prompts, i),
                system_prompt=_get_helper(system_prompts, i),
                temperature=_get_helper(temperatures, i),
                max_tokens=_get_helper(max_tokens, i),
            )
            for i in range(n)
        ]
        return await asyncio.gather(*calls)

    async def _call_wrapper(self, **kwargs: Any) -> str:
        return self.call(**kwargs)

    async def _call_custom_wrapper(self, **kwargs: Any) -> str:
        return self.call_custom(**kwargs)


def _get_helper(param: Optional[List[Any]], i: int) -> Optional[Any]:
    if param is None:
        return None
    if len(param) == 1:
        return param[0]
    return param[i]


def _check_concurrent_params(
    n: int,
    params: List[Optional[List[Any]]],
    argnames: List[str],
) -> None:
    if n < 1:
        raise ValueError("n must be at least 1")

    for i, param in enumerate(params):
        if param is not None and len(param) not in [1, n]:
            raise ValueError(
                f"{argnames[i]} must have length 1 or {n}, got {len(param)}"
            )
